#!/usr/bin/env python3
import enum
import numpy as np
import urllib.request as req
import math
import collections
import sys
import readline
import curses

class Moves(enum.Enum):
	left = "L"
	right = "p"
	up = "N"
	down = "D"

class Player:
	def __init__(self, xpos, ypos, treasure, score):
		self.xpos = xpos
		self.ypos = ypos
		self.treasure = treasure
		self.score = score

class Server:
	GameEnd = collections.namedtuple("GameEnd", ["score", "escore"])

	class Response:
		def __init__(self):
			self.board = None
			self.player = None
			self.eplayer = None

	def __init__(self, url):
		self.url = url
		self.id = None
		self.moves_left = None
		
	def parse_splitin(self, splitin):
		response = Server.Response()

		response.board = Board("".join(splitin[2:12]))

		playerinfo = splitin[12].split()
		response.player = Player(int(playerinfo[1]) - 1,
			int(playerinfo[0]) - 1, int(playerinfo[2]),
			int(playerinfo[3]))
		if(len(splitin) > 14): # Hra s dvoumi hráči
			eplayerinfo = splitin[13].split()
			response.eplayer = Player(int(eplayerinfo[1]) - 1,
				int(eplayerinfo[0]) - 1, int(eplayerinfo[2]),
				int(eplayerinfo[3]))

		return response

	def start(self):
		resp = req.urlopen(self.url + "/start")
		text = resp.read().replace(b"\r\n", b"\n").decode()
		splitin = text.splitlines()
		self.id = int(splitin[0])
		self.moves_left = int(splitin[1])
		
		return self.parse_splitin(splitin)
	
	def submit_move(self, move):
		resp = req.urlopen(self.url + "/tah?id=" + str(self.id) + "&pohyb=" + move)
		text = resp.read().replace(b"\r\n", b"\n").decode()
		
		if text.startswith("Konec hry:"):
			splitin = text.split()
			if len(splitin) == 3: # Hra s jedním hráčem
				return Server.GameEnd(int(splitin[2]), None)
			else: #Hra dvou hráčů
				return Server.GameEnd(int(splitin[2]), int(splitin[3]))
		
		if text.startswith("Chyba:"):
			splitin = text.split()
			print("Došlo k chybě:", splitin[1])
			sys.exit(1)

		splitin = text.splitlines()
		self.moves_left = int(splitin[1])
		return self.parse_splitin(splitin)
	
class BPieces(enum.Enum):
	space = ".."
	wall = "XX"
	base = "++"
	ebase = "--"

class Board:
	def __init__(self, text):
		"""
		Inicializuje hrací pole z textu
		text by měl být string o 20*10 znaků (žádné newliny)
		"""
		self.board = np.zeros((10, 10), dtype=object)
		for i in range(10*10):
			fstr =  text[i*2:i*2 + 2]
			if fstr.isdigit():
				fval = int(fstr)
			else:
				fval = fstr

			self.set_piece(math.floor(i/10), i%10, fstr)

	def set_piece(self, x, y, piece):
		self.board[x][y] = piece
	
	def get_piece(self, x, y):
		return self.board[x][y]

	def render(self):
		"""
		Vykreslí hrací pole na stdout
		"""
		#Čára nahoře
		game.stdscr.move(6, 0)
		for i in range(self.board.shape[0] * 3 + 1):
			game.stdscr.addstr("-", 0)
		game.stdscr.move(game.stdscr.getyx()[0] + 1, 0)
		
		i = 0
		for col in self.board:
			j = 0
			game.stdscr.addstr("|", 0)
			for piece in col:
				if (i, j) == (game.player.xpos, game.player.ypos):
					game.stdscr.addstr(str(piece), curses.color_pair(1))
					game.stdscr.addstr("|")
				elif game.eplayer != None and\
					(i,j) == (game.eplayer.xpos, game.eplayer.ypos):
					game.stdscr.addstr(str(piece), curses.color_pair(2))
					game.stdscr.addstr("|")
				else:
					game.stdscr.addstr(str(piece) + "|", 0)
				j += 1
			game.stdscr.move(game.stdscr.getyx()[0] + 1, 0)
			i += 1
			
		
		#Čára dole
		for i in range(self.board.shape[0] * 3 + 1):
			game.stdscr.addstr("-", 0)

#Dictionary, který mapuje vstup z klávesnice na příkazy pro server
commands = {
	"h": Moves.left, "a": Moves.left,
	"j": Moves.down, "s": Moves.down,
	"k": Moves.up, "w": Moves.up,
	"l": Moves.right, "d": Moves.right
}

class Game:
	def __init__(self):
		if len(sys.argv) != 2:
			print("Syntaxe:", sys.argv[0], "server URL")
			print("např.", sys.argv[0], "http://localhost:8000")
			sys.exit(1)

		self.server = Server(sys.argv[1])
		response = self.server.start()
		self.board = response.board
		self.player = response.player
		self.eplayer = response.eplayer

		self.stdscr = curses.initscr()
		curses.noecho()
		curses.cbreak()
		self.stdscr.keypad(True)
		curses.curs_set(False)

		curses.start_color()
		
		#Barvičky!
		curses.init_pair(1, 0, curses.COLOR_YELLOW)
		curses.init_pair(2, 0, curses.COLOR_RED)

	def end(self):
		self.stdscr.clear()
		curses.nocbreak()
		self.stdscr.keypad(False)
		curses.echo()
		curses.curs_set(True)
		curses.endwin()

	def run(self):
		while True:
			self.stdscr.addstr(0, 0, "Lovci Pokladů (David Koňařík)", 0)
			self.stdscr.addstr(1, 0, "Tato hra funguje nejlépe v terminálech " +
				"kompatibilních s ncurses a barvami", 0)
			self.stdscr.addstr(2, 0, "Ovládání: hjkl/wasd - pohyb, q - ukončit", 0)
	
			self.stdscr.addstr(3, 0, "Poklad nyní: " + str(self.player.treasure) +
				" Poklad celkem: " +  str(self.player.score), 0)
			self.stdscr.addstr(4, 0, "Tahů do konce hry: " +  str(self.server.moves_left), 0)
			self.board.render()
			self.stdscr.refresh()
			incommand = chr(self.stdscr.getch())
			self.stdscr.clear()

			if incommand == "q":
				self.end()
				sys.exit(0)
			
			if incommand not in commands:
				self.stdscr.addstr(5, 0, "Špatný příkaz", 0)
				continue
			
			command = commands[incommand].value.upper()

			command_dirs = {"L": (0, -1), "P": (0, 1), "N": (-1, 0),
				"D": (1, 0)}

			nx = self.player.xpos + command_dirs[command][0]
			ny = self.player.ypos + command_dirs[command][1]
			
			if nx >= 10 or ny >= 10 or nx < 0 or ny < 0:
				self.stdscr.addstr(5, 0, "Neplatný tah", 0)
				continue

			npiece = self.board.get_piece(nx, ny)
			if npiece == BPieces.wall.value or npiece == BPieces.ebase.value or\
				(self.eplayer != None and self.eplayer.xpos == nx and
				self.eplayer.ypos == ny):
				self.stdscr.addstr(5, 0, "Neplatný tah", 0)
				continue

			result = self.server.submit_move(command)
			if isinstance(result, Server.GameEnd):
				self.end()
				if self.eplayer == None:
					print("Hra skončila. Konečné skóre bylo:", result.score)
				else:
					print("Hra skončila. Vaše konečné skóre bylo:", result.score)
					print("Konečné skóre protivníka bylo:", result.escore)
				sys.exit(0)
			else:
				self.board = result.board
				self.player = result.player
				self.eplayer = result.eplayer

game = Game()
game.run()

